package kmsHandler

import (
	"fmt"

	"code.fbi.h-da.de/danet/gnmi-target/handler"
	"code.fbi.h-da.de/danet/quant/goKMS/kms"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/event"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/util"
	gnmitargetygot "code.fbi.h-da.de/danet/quant/goKMS/model"
	"github.com/openconfig/gnmi/proto/gnmi"
	"github.com/openconfig/ygot/ygot"
	log "github.com/sirupsen/logrus"
)

type KeyRoutingSessionHandler struct {
	handler.DefaultPathHandler
	kms    *kms.KMS
	events <-chan event.Event
}

func NewKeyRoutingSessionHandler(kms *kms.KMS) *KeyRoutingSessionHandler {
	return &KeyRoutingSessionHandler{
		DefaultPathHandler: handler.DefaultPathHandler{
			Name: "kms-key-routing-session-handler",
			Paths: map[string]struct{}{
				"/key-routing-sessions": {},
			},
		},
		kms: kms,
	}
}

func (yh *KeyRoutingSessionHandler) Init(config *handler.Config, publishToSubsFunc func([]*gnmi.Notification) error) error {
	yh.Config = config
	yh.PublishToSubs = publishToSubsFunc

	var err error
	yh.events, err = yh.kms.EventBus().Subscribe(event.ROUTE)
	if err != nil {
		return err
	}

	// Fill out the known fields in the ygot structs
	_, err = yh.updateOrCreateKeyRoutingSessions(yh.kms)
	if err != nil {
		return err
	}

	// Start the go routine that takes care of any update from the kms
	go func() {
		for {
			select {
			case <-yh.events:
				log.Println("Update for Routes.")
				diff, err := yh.updateOrCreateKeyRoutingSessions(yh.kms)
				if err != nil {
					log.Errorf("Error within interface subscription goroutine; %v", err)
					break
				}
				if err := yh.PublishToSubs(diff); err != nil {
					log.Errorf("Error within interface subscription goroutine; %v", err)
				}
			}
		}
	}()

	return nil
}

func (yh *KeyRoutingSessionHandler) Update(c ygot.ValidatedGoStruct, jobs []*gnmi.Update) error {
	fmt.Println("Update request received for ", yh.Name)
	config, ok := c.(*gnmitargetygot.Gnmitarget)
	if !ok {
		return fmt.Errorf("failed type assertion for config %T", (*gnmitargetygot.Gnmitarget)(nil))
	}

	// The changedPathId map is used to collect all the path-id keys from jobs.
	// Since jobs are all the gnmi updates from a ygot.Diff there can be
	// multiple entries for the same path-id as key. Therefore the
	// map["pathId"]struct{} is used to easily avoid duplicates.
	changedPathIds := make(map[string]struct{})

	for _, job := range jobs {
		pathElements := job.GetPath().GetElem()
		if len(pathElements) >= 2 {
			pathId, ok := pathElements[1].GetKey()["path-id"]
			if ok {
				changedPathIds[pathId] = struct{}{}
			} else {
				log.Errorf("Path does not contain path-id key")
			}
		} else {
			log.Errorf("Path elements are less than 2")
		}
	}

	log.Debugf("KeyRoutingSessionHandler: Update triggered for the following path-Ids, %v", changedPathIds)

	routingSessions := config.GetOrCreateKeyRoutingSessions()
	for pathId := range changedPathIds {
		routingSession := routingSessions.GetRoutingSessions(pathId)
		pathId := routingSession.GetPathId()
		nextHop := routingSession.GetNextHop()
		prevHop := routingSession.GetPrevHop()

		var nextHopString string
		var prevHopString string

		if nextHop != nil {
			nextHopString = nextHop.GetNodeId()
		}
		if prevHop != nil {
			prevHopString = prevHop.GetNodeId()
		}

		var initKMS *util.RemoteKMS
		// check if initiating kms address is set
		if initiatingKmsAddress := routingSession.GetInitiatingKmsAddress(); initiatingKmsAddress != nil {
			addressPrefix := ""
			if initiatingKmsAddress.GetIpAddress() != "" {
				addressPrefix = initiatingKmsAddress.GetIpAddress()
			} else if initiatingKmsAddress.GetHostname() != "" {
				addressPrefix = initiatingKmsAddress.GetHostname()
			} else {
				log.Error("initiatingKmsAddress must have either an IP address or a hostname")
				return fmt.Errorf("initiatingKmsAddress must have either an IP address or a hostname")
			}

			initKMS = &util.RemoteKMS{
				Id:      initiatingKmsAddress.GetNodeId(),
				Address: addressPrefix,
				Port:    initiatingKmsAddress.GetPort(),
			}
		}

		err := yh.kms.AssignForwardingRoute(pathId, prevHopString, nextHopString, initKMS)
		if err != nil {
			return err
		}
	}

	return nil
}

func (yh *KeyRoutingSessionHandler) updateOrCreateKeyRoutingSessions(kms *kms.KMS) ([]*gnmi.Notification, error) {
	yh.Config.Lock()
	defer yh.Config.Unlock()

	copyCurrentConfig, err := ygot.DeepCopy(yh.Config.Data)
	if err != nil {
		return nil, err
	}

	newConfig, ok := copyCurrentConfig.(*gnmitargetygot.Gnmitarget)
	if !ok {
		return nil, fmt.Errorf("Wrong type, exptected: %T, got: %T", (*gnmitargetygot.OpenconfigInterfaces_Interfaces)(nil), copyCurrentConfig)
	}

	confKeyRoutingSessions := newConfig.GetOrCreateKeyRoutingSessions()

	kmsRoutingTable := kms.RoutingTableDeepCopy()

	for _, route := range kmsRoutingTable {
		confTempRoutingSession := confKeyRoutingSessions.GetOrCreateRoutingSessions(route.PathId.String())

		if route.Previous != nil {
			confPrevHop := confTempRoutingSession.GetOrCreatePrevHop()
			confPrevHop.NodeId = ygot.String(route.Previous.GetKmsPeerId().String())
			confPrevHop.IpAddress = ygot.String(route.Previous.Address().IP.String())
			confPrevHop.Port = ygot.Uint16(route.Previous.Address().AddrPort().Port())
		}

		if route.Next != nil {
			confNextHop := confTempRoutingSession.GetOrCreateNextHop()
			confNextHop.NodeId = ygot.String(route.Next.GetKmsPeerId().String())
			confNextHop.IpAddress = ygot.String(route.Next.Address().IP.String())
			confNextHop.Port = ygot.Uint16(route.Next.Address().AddrPort().Port())
		}

		if route.RemoteKMS != nil {
			confTempRoutingSessionInitAddress := confTempRoutingSession.GetOrCreateInitiatingKmsAddress()
			if route.RemoteKMS.Address != "" {
				confTempRoutingSessionInitAddress.IpAddress = ygot.String(route.RemoteKMS.Address)
				confTempRoutingSessionInitAddress.Port = ygot.Uint16(route.RemoteKMS.Port)
				confTempRoutingSessionInitAddress.Hostname = ygot.String(route.RemoteKMS.Address) // TODO: get real hostname here
			}
			if route.RemoteKMS.Id != "" {
				confTempRoutingSessionInitAddress.NodeId = ygot.String(route.RemoteKMS.Id)
			}
		}

		// TODO: add key properties
		// confKeyProperties := confTempRoutingSession.GetOrCreateKeyProperties()

		// TODO: add amount key forwarded
		// confTempRoutingSession.AmountKeysForwarded = ygot.Int64(0)

		// TODO: add operation status
	}

	//validate struct
	if err := newConfig.Validate(); err != nil {
		return nil, err
	}

	notifications, err := ygot.DiffWithAtomic(yh.Config.Data, newConfig)
	if err != nil {
		return nil, err
	}

	yh.Config.Data = newConfig

	return notifications, nil
}
