package kmsHandler

import (
	"fmt"

	"code.fbi.h-da.de/danet/gnmi-target/handler"
	"code.fbi.h-da.de/danet/quant/goKMS/kms"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/event"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/peers"
	gnmitargetygot "code.fbi.h-da.de/danet/quant/goKMS/model"
	"github.com/openconfig/gnmi/proto/gnmi"
	"github.com/openconfig/ygot/ygot"
	log "github.com/sirupsen/logrus"
)

type PeerHandler struct {
	handler.DefaultPathHandler
	events <-chan event.Event
	kms    *kms.KMS
}

func NewPeerHandler(kms *kms.KMS) *PeerHandler {
	return &PeerHandler{
		DefaultPathHandler: handler.DefaultPathHandler{
			Name: "kms-peer-handler",
			Paths: map[string]struct{}{
				"/kms-peers": {},
			},
		},
		kms: kms,
	}
}

func (yh *PeerHandler) Init(config *handler.Config, publishToSubsFunc func([]*gnmi.Notification) error) error {
	yh.Config = config
	yh.PublishToSubs = publishToSubsFunc

	var err error
	yh.events, err = yh.kms.EventBus().Subscribe(event.PEER)
	if err != nil {
		return err
	}

	// Fill out the known fields in the ygot structs
	_, err = yh.updateOrCreatePeerTable(yh.kms)
	if err != nil {
		return err
	}

	// Start the go routine that takes care of any update from the kms
	go func() {
		for {
			select {
			case <-yh.events:
				log.Println("Update for KMSPeer.")
				//lock access for model
				diff, err := yh.updateOrCreatePeerTable(yh.kms)
				if err != nil {
					log.Errorf("Error within interface subscription goroutine; %v", err)
					break
				}
				if err := yh.PublishToSubs(diff); err != nil {
					log.Errorf("Error within interface subscription goroutine; %v", err)
				}
			}
		}
	}()

	return nil
}

func (yh *PeerHandler) Update(c ygot.ValidatedGoStruct, jobs []*gnmi.Update) error {
	return nil
}

func (yh *PeerHandler) updateOrCreatePeerTable(kms *kms.KMS) ([]*gnmi.Notification, error) {
	yh.Config.Lock()
	defer yh.Config.Unlock()

	copyCurrentConfig, err := ygot.DeepCopy(yh.Config.Data)
	if err != nil {
		return nil, err
	}

	newConfig, ok := copyCurrentConfig.(*gnmitargetygot.Gnmitarget)
	if !ok {
		return nil, fmt.Errorf("Wrong type, exptected: %T, got: %T", (*gnmitargetygot.OpenconfigInterfaces_Interfaces)(nil), copyCurrentConfig)
	}

	// Create ygot structs for the kms
	confKmsPeerTable := newConfig.GetOrCreateKmsPeerTable()

	kmsPeers := kms.PeersDeepCopy()

	for _, peer := range kmsPeers {
		confTempKmsPeer := confKmsPeerTable.GetOrCreateKmsPeers(peer.GetKmsPeerId().String())
		confTempPeerInformation := confTempKmsPeer.GetOrCreatePeerInformation()
		confTempPeerAddress := confTempPeerInformation.GetOrCreatePeerAddress()

		confTempPeerAddress.IpAddress = ygot.String(peer.Address().IP.String())
		confTempPeerAddress.Port = ygot.Uint16(uint16(peer.Address().Port))
		confTempPeerAddress.Hostname = ygot.String(peer.Address().String()) // TODO: get real hostname here
		confTempPeerAddress.NodeId = ygot.String(peer.GetKmsPeerId().String())

		client := peer.Client()
		switch {
		case client == nil:
			log.Error("gRPC client for internal peer communication is not set.")
		case client.KmsTalkerClient != nil:
			confInterKmsProtocol := confTempPeerInformation.GetOrCreateInterKmsProtocol()
			confInterKmsProtocol.ProtocolName = gnmitargetygot.Temp_KmsInterfaceTable_KmsInterfaces_ProtocolName_DANET_INTER_KMS_PROTO
		default:
			log.Error("unknown client for internal peer communication.")
		}

		confTempPeerInformation.QkdModuleId = ygot.String(peer.QuantumModule().ID().String())

		switch status := peer.Status(); status {
		case peers.KmsPeerUp:
			confTempPeerInformation.OpStatus = gnmitargetygot.Temp_OpStatus_ESTABLISHED
		case peers.KmsPeerPending:
			confTempPeerInformation.OpStatus = gnmitargetygot.Temp_OpStatus_PENDING
		default:
			confTempPeerInformation.OpStatus = gnmitargetygot.Temp_OpStatus_ERROR
		}

		// NOTE: Currently it is  assumed that only 256 bit keys are necessary.
		confTempPeerInformation.NegotiatedKeyLength = ygot.Uint64(256)
	}

	//validate struct
	if err := newConfig.Validate(); err != nil {
		return nil, err
	}

	notifications, err := ygot.DiffWithAtomic(yh.Config.Data, newConfig)
	if err != nil {
		return nil, err
	}

	yh.Config.Data = newConfig

	return notifications, nil
}
