package etsi14Server

import (
	"context"
	"errors"
	"fmt"
	"net/http"
	"os"
	"os/signal"
	"sync"
	"syscall"
	"time"

	restserver "code.fbi.h-da.de/danet/quant/etsi014/go/rest/etsi/server/go"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/crypto"
	"github.com/google/uuid"
	"github.com/sirupsen/logrus"
)

const etsi014RequestID = "etsi014"

type ETSI14RESTService struct {
	serviceAddress  string
	remoteCKMSID    uuid.UUID
	keyStoreMutex   sync.RWMutex
	keyStore        map[string]string
	KeyStoreChannel chan []crypto.KSAKey
	httpServer      *http.Server
	forwardingFunc  func(uuid.UUID, int64, string) ([]crypto.KSAKey, error)
}

func NewETSI14RESTService(serviceAddress string, remoteCKMSID uuid.UUID, forwardingFunc func(uuid.UUID, int64, string) ([]crypto.KSAKey, error)) *ETSI14RESTService {
	return &ETSI14RESTService{
		serviceAddress:  serviceAddress,
		remoteCKMSID:    remoteCKMSID,
		keyStore:        make(map[string]string, 0),
		KeyStoreChannel: make(chan []crypto.KSAKey, 10),
		forwardingFunc:  forwardingFunc,
	}
}

// Run initiates all the necessities and starts the etsi14 server.
func (s *ETSI14RESTService) Run() {
	logrus.Info("Starting ETSI14 server...")

	etsi14APIController := restserver.NewDefaultAPIController(s)

	router := restserver.NewRouter(etsi14APIController)

	s.httpServer = &http.Server{
		Addr:              s.serviceAddress,
		Handler:           router,
		ReadHeaderTimeout: 15 * time.Second,
		ReadTimeout:       15 * time.Second,
		WriteTimeout:      10 * time.Second,
		IdleTimeout:       30 * time.Second,
	}

	go s.startServer()

	ctx, cancel := context.WithCancel(context.Background())
	stopChannel := make(chan os.Signal, 1)
	signal.Notify(stopChannel, syscall.SIGINT)

	defer func() {
		logrus.Info("Shutting down ETSI14 server...")
		if err := s.httpServer.Shutdown(ctx); err != nil {
			logrus.Errorf("Error shutting down server: %v", err)
			_ = s.httpServer.Close()
			return
		}
		logrus.Info("Gracefully shutdown server completed.")
	}()

	logrus.Infof("ETSI14 server running on: %s", s.serviceAddress)

	go s.runKeyStoreChannel()

	<-stopChannel

	cancel()
}

// startServer starts the listener of the etsi14 HTTP server.
func (s *ETSI14RESTService) startServer() {
	if err := s.httpServer.ListenAndServe(); !errors.Is(err, http.ErrServerClosed) {
		logrus.Error(err)
	}

	logrus.Info("Stopped serving new connections.")
}

// runKeyStoreChannel sets up a listener for when a key needs to be added to the servers key store.
func (s *ETSI14RESTService) runKeyStoreChannel() {
	for {
		select {
		case keys := <-s.KeyStoreChannel:
			logrus.Debugf("(ETSI14) KeyStoreChannel received: %v", keys)
			s.pushKeysToStoreAfterExchange(keys)
		}
	}
}

// GetKey returns a number of keys of specified size and safes the slaveSAEID for further use.
func (s *ETSI14RESTService) GetKey(ctx context.Context, slaveSAEID string, number int64, size int64) (restserver.ImplResponse, error) {
	keyContainer, err := s.getKeyContainerFromKeyExchange(number)
	if err != nil {
		return restserver.Response(http.StatusInternalServerError, keyContainer), err
	}

	logrus.Debugf("(ETSI14) GetKey: Sending %d keys, Incoming ctx: %s", number, ctx)

	return restserver.Response(http.StatusOK, keyContainer), nil
}

// GetKeyPost returns a number of keys of specified size and safes the slaveSAEID for further use.
func (s *ETSI14RESTService) GetKeyPost(ctx context.Context, slaveSAEID string, keyRequest restserver.KeyRequest) (restserver.ImplResponse, error) {
	if keyRequest.Number < 1 {
		logrus.Errorf("(ETSI14) number must be positive and at least 1, provided: %d\n", keyRequest.Number)
		return restserver.Response(http.StatusBadRequest, nil), fmt.Errorf("requested invalid number of keys")
	}

	keyContainer, err := s.getKeyContainerFromKeyExchange(keyRequest.Number)
	if err != nil {
		return restserver.Response(http.StatusInternalServerError, keyContainer), err
	}

	logrus.Debugf("(ETSI14) GetKey: Sending %d keys, Incoming ctx: %s", keyRequest.Number, ctx)

	return restserver.Response(http.StatusOK, keyContainer), nil
}

// GetKeyWithIds returns a key with a specific ID related to the masterSAEID.
func (s *ETSI14RESTService) GetKeyWithIds(ctx context.Context, masterSAEID string, keyID string) (restserver.ImplResponse, error) {
	s.keyStoreMutex.RLock()
	key, ok := s.keyStore[keyID]
	if !ok {
		logrus.Errorf("(ETSI14) Key with ID %s not found", keyID)
		return restserver.Response(http.StatusBadRequest, nil), nil
	}
	s.keyStoreMutex.RUnlock()

	keyContainer := &restserver.KeyContainer{
		Keys: []restserver.KeyContainerKeysInner{
			{
				KeyID: keyID,
				Key:   key,
			},
		},
	}

	logrus.Debugf("(ETSI14) GetKeyWithIds: Sending key with ID: %s, Incoming ctx: %s", keyID, ctx)

	return restserver.Response(http.StatusOK, keyContainer), nil
}

// GetKeyWithIdsPost returns a key with a specific ID related to the masterSAEID.
func (s *ETSI14RESTService) GetKeyWithIdsPost(ctx context.Context, masterSAEID string, keyIdsRequest restserver.KeyIdsRequest) (restserver.ImplResponse, error) {
	if keyIdsRequest.KeyIDs == nil {
		logrus.Error("(ETSI14) GetKeyWithIdsPost: Empty KeyIdsRequest.KeyIDs")
		return restserver.Response(http.StatusBadRequest, nil), nil
	}

	s.keyStoreMutex.RLock()
	key, ok := s.keyStore[keyIdsRequest.KeyIDs[0].KeyID]
	if !ok {
		logrus.Errorf("(ETSI14) Key with ID %s not found", keyIdsRequest.KeyIDs[0].KeyID)
		logrus.Debugf("Current store: %v", s.keyStore)
		return restserver.Response(http.StatusBadRequest, nil), nil
	}
	s.keyStoreMutex.RUnlock()

	keyContainer := &restserver.KeyContainer{
		Keys: []restserver.KeyContainerKeysInner{
			{
				KeyID: keyIdsRequest.KeyIDs[0].KeyID,
				Key:   key,
			},
		},
	}

	logrus.Debugf("(ETSI14) GetKeyWithIdsPost: Sending key with ID: %s, Incoming ctx: %s", keyIdsRequest.KeyIDs[0].KeyID, ctx)

	return restserver.Response(http.StatusOK, keyContainer), nil
}

// GetStatus returns status information of this network element.
func (s *ETSI14RESTService) GetStatus(ctx context.Context, slaveSAEID string) (restserver.ImplResponse, error) {
	status := restserver.Status{
		SourceKMEID:      "AAAABBBBCCCCDDDD",
		TargetKMEID:      "EEEEFFFFGGGGHHHH",
		MasterSAEID:      "IIIIJJJJKKKKLLLL",
		SlaveSAEID:       "MMMMNNNNOOOOPPPP",
		KeySize:          352,
		StoredKeyCount:   25000,
		MaxKeyCount:      100000,
		MaxKeyPerRequest: 128,
		MaxKeySize:       1024,
		MinKeySize:       64,
		MaxSAEIDCount:    0,
	}

	return restserver.Response(http.StatusNotImplemented, status), nil
}

// getKeyContainerFromKeyExchange starts the key exchange using the underlying QKD forwarding and returns the resulting KSA keys.
func (s *ETSI14RESTService) getKeyContainerFromKeyExchange(number int64) (*restserver.KeyContainer, error) {
	keys, err := s.forwardingFunc(s.remoteCKMSID, number, etsi014RequestID) // Note: we should have a mapping mechanism here to support multiple endpoints in the future
	if err != nil {
		return nil, err
	}

	keyContainer := &restserver.KeyContainer{}
	for _, key := range keys {
		keyContainer.Keys = append(keyContainer.Keys,
			restserver.KeyContainerKeysInner{
				KeyID: key.KeyID,
				Key:   key.Key,
			})

		s.pushToKeyStore(key)

		logrus.Debugf("(ETSI14) GetKey: Adding key with ID: %s to store", key.KeyID)
	}

	return keyContainer, nil
}

// pushKeysToStoreAfterExchange pushes all the KSA keys received on server side to the local key store.
func (s *ETSI14RESTService) pushKeysToStoreAfterExchange(keys []crypto.KSAKey) {
	for _, key := range keys {
		s.pushToKeyStore(key)
	}
}

// pushToKeyStore pushes one KSA key to the local key store.
func (s *ETSI14RESTService) pushToKeyStore(key crypto.KSAKey) {
	s.keyStoreMutex.Lock()
	defer s.keyStoreMutex.Unlock()

	s.keyStore[key.KeyID] = key.Key

	logrus.Debugf("(ETSI14) Pushed key with ID: %s to keystore", key.KeyID)
}
