package kms

import (
	"context"
	"net"
	"strconv"
	"time"

	"code.fbi.h-da.de/danet/quant/goKMS/kms/peers"
	"code.fbi.h-da.de/danet/quant/quantumlayer"
	pb "code.fbi.h-da.de/danet/quipsec/gen/go/quipsec"
	"github.com/sirupsen/logrus"
	"google.golang.org/grpc/codes"
	"google.golang.org/grpc/peer"
	"google.golang.org/grpc/status"
)

type quipSecServer struct {
	pb.UnimplementedKmsQkdmCommunicationServiceServer
	KMS *KMS
}

func (qs *quipSecServer) PushKeys(ctx context.Context, req *pb.PushKeysRequest) (*pb.PushKeysResponse, error) {
	p, _ := peer.FromContext(ctx)
	host, _, err := net.SplitHostPort(p.Addr.String())
	if err != nil {
		logrus.Errorf("could not get host from peer")
		return nil, status.Errorf(codes.Internal, "could not get host from peer")
	}
	logrus.Infof("Got host: %s", host)

	qs.KMS.quantumModulesMutex.RLock()
	qm, ok := qs.KMS.quantumModules[host]
	qs.KMS.quantumModulesMutex.RUnlock()
	if !ok {
		return nil, status.Errorf(codes.Internal, "could not find a quantum module for host address: %s", host)
	}

	eqm, ok := qm.(*peers.DanetQuantumModule)
	if !ok {
		logrus.Errorf("quantum module is of wrong type")
		return nil, status.Errorf(codes.Internal, "quantum module is of wrong type")
	}

	bulkKeyId, err := strconv.ParseInt(req.GetKeyBulk().GetKeyId(), 10, 64)
	if err != nil {
		logrus.Errorf("could not get bulkkeyid")
		return nil, status.Errorf(codes.Internal, "could not get bulkkeyid")
	}

	if !eqm.IsActive() {
		logrus.Debugf("Quantum module: %s is not active and denied incoming bulk keys", eqm.ID())
		return nil, status.Errorf(codes.Aborted, "Currently no new bulk keys are accepted")
	}
	eqm.RawBulkKeysMutex.Lock()
	eqm.RawBulkKeys[bulkKeyId] = &quantumlayer.QuantumLayerBulkKey{
		BulkKeyId:     bulkKeyId,
		BulkKeyLength: int(req.GetKeyBulk().GetKeyLength()),
		BulkKey:       req.GetKeyBulk().Keys,
	}
	eqm.RawBulkKeysMutex.Unlock()

	if eqm.KeyStore().Length() == 0 {
		if err := eqm.Sync(); err != nil {
			logrus.Error("Syncing key stores after a bulk key push failed; err: ", err)
		}
	}

	logrus.Debugf("%s received a new bulk from: %s with id: %s and a length of: %d", qs.KMS.kmsName, qm.Address(), req.GetKeyBulk().GetKeyId(), req.GetKeyBulk().GetKeyLength())
	return &pb.PushKeysResponse{Timestamp: time.Now().Unix()}, nil
}
