package integration_tests

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io"
	"net/http"
	"os"
	"testing"
	"time"

	utils "code.fbi.h-da.de/danet/quant/integration-tests/code/integrationTestUtils"
	"github.com/stretchr/testify/assert"
)

type ETSI14RequestData struct {
	Number int64                      `json:"number,omitempty"`
	KeyIDs []KeyIdsRequestKeyIdsInner `json:"key_IDs,omitempty"`
}

type KeyIdsRequestKeyIdsInner struct {
	KeyID string `json:"key_ID"`
}

type ETSI14ResponseData struct {
	Keys []KeyContainerKeysInner `json:"Keys,omitempty"`
}

type KeyContainerKeysInner struct {
	KeyID string `json:"key_ID,omitempty"`

	Key string `json:"key,omitempty"`
}

func getKeyContainerFromResponse(kmsResp *http.Response) (*ETSI14ResponseData, error) {
	defer kmsResp.Body.Close() //nolint:errcheck
	body, err := io.ReadAll(kmsResp.Body)
	if err != nil {
		return nil, err
	}

	kmsKeyContainer := &ETSI14ResponseData{}
	err = json.Unmarshal([]byte(body), kmsKeyContainer)
	if err != nil {
		return nil, err
	}

	return kmsKeyContainer, nil
}

func TestETSI14GetKey(t *testing.T) { //nolint:gocyclo
	gosdncScript := "../../config/gosdnc/add_devices.sh"
	gosdncScript_ENV := os.Getenv("INTEGRATION_TEST_GOSDNC_SCRIPT")
	if gosdncScript_ENV != "" {
		gosdncScript = gosdncScript_ENV
	}
	cleanupScript := "../../config/gosdnc/delete_devices.sh"
	cleanupScript_ENV := os.Getenv("INTEGRATION_TEST_CLEANUP_SCRIPT")
	if cleanupScript_ENV != "" {
		cleanupScript = cleanupScript_ENV
	}
	controllerURL := "127.0.0.1:55055"
	controllerURL_ENV := os.Getenv("INTEGRATION_TEST_CONTROLLER_URL")
	if controllerURL_ENV != "" {
		controllerURL = controllerURL_ENV
	}
	kms1ETSI14URL := "127.0.0.1:1414"
	kms1ETSI14URL_ENV := os.Getenv("INTEGRATION_TEST_KMS1_ETSI14_URL")
	if kms1ETSI14URL_ENV != "" {
		kms1ETSI14URL = kms1ETSI14URL_ENV
	}
	kms2ETSI14URL := "127.0.0.1:1415"
	kms2ETSI14URL_ENV := os.Getenv("INTEGRATION_TEST_KMS2_ETSI14_URL")
	if kms2ETSI14URL_ENV != "" {
		kms2ETSI14URL = kms2ETSI14URL_ENV
	}

	// Tell the qkdn-controller what devices to use.
	_, err := utils.RunGosdncScript(gosdncScript, controllerURL)
	if err != nil {
		currentFolderPath := os.Getenv("PWD")
		t.Errorf("Error running gosdnc script. Current folder path: %s, Error: %s", currentFolderPath, err)
	}
	defer utils.RunGosdncScript(cleanupScript, controllerURL) //nolint:errcheck

	url := fmt.Sprintf("http://%s/api/v1/keys/slave_SAE_ID/enc_keys", kms1ETSI14URL)
	data := ETSI14RequestData{
		Number: 1,
	}

	jsonData, err := json.Marshal(data)
	if err != nil {
		fmt.Println(err)
		return
	}

	var respKms1 *http.Response
	go func() {
		respKms1, err = http.Post(url, "application/json", bytes.NewBuffer(jsonData))
		if err != nil {
			t.Errorf("Error making HTTP request: %s", err)
			return
		}

		if respKms1.StatusCode != http.StatusOK {
			t.Errorf("Expected status code 200 OK, but got %d", respKms1.StatusCode)
		}
	}()

	time.Sleep(time.Duration(2) * time.Second)

	kms1KeyContainer, err := getKeyContainerFromResponse(respKms1)
	if err != nil {
		t.Errorf("Error getting Key Container from resp: %v, err: %v", respKms1, err)
	}

	assert.NotNil(t, kms1KeyContainer.Keys)

	url = fmt.Sprintf("http://%s/api/v1/keys/master_SAE_ID/dec_keys", kms2ETSI14URL)
	data = ETSI14RequestData{
		KeyIDs: []KeyIdsRequestKeyIdsInner{
			{
				KeyID: kms1KeyContainer.Keys[0].KeyID,
			},
		},
	}

	jsonData, err = json.Marshal(data)
	if err != nil {
		fmt.Println(err)
		return
	}

	var respKms2 *http.Response
	go func() {
		respKms2, err = http.Post(url, "application/json", bytes.NewBuffer(jsonData))
		if err != nil {
			t.Errorf("Error making HTTP request: %s", err)
			return
		}

		if respKms2.StatusCode != http.StatusOK {
			t.Errorf("Expected status code 200 OK, but got %d", respKms2.StatusCode)
		}
	}()

	time.Sleep(time.Duration(2) * time.Second)

	kms2KeyContainer, err := getKeyContainerFromResponse(respKms2)
	if err != nil {
		t.Errorf("Error getting Key Container from resp: %v, err: %v", respKms1, err)
	}

	assert.NotNil(t, kms2KeyContainer.Keys)

	assert.Equal(t, kms1KeyContainer, kms2KeyContainer)
}
